/**
 This file is software from the ORKWAN project.
 http://www.orkwan.com
 
 Copyright (C) 2000-2008 Henrik Nyberg.
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "ORKWAN" must not be used to endorse or promote products
    derived from this software without prior written permission.  For
    written permission, please contact <info_AT_orkwan_DOT_com>.

 4. Products derived from this software may not be called "ORKWAN", nor
    may "ORKWAN" appear in their name, without prior written permission
    from the ORKWAN Project <info_AT_orkwan_DOT_com>.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:
     "This product includes software developed by the
      ORKWAN Project (http://www.orkwan.com/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
 */

package com.orkwan;

import org.jdom.Document;
import org.jdom.Element;

import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.util.HashMap;

/**
 * OrkwanController is the abstract base class for all controllers in the Orkwan MVC framework.
 * Inherit from this class to create a base class controller for a web application.
 * Group common functionality for the web application in the base class controller.
 * Further subclass the application base class controller to create web functions for the application. 
 *
 * @author Henrik Nyberg
 */
public abstract class OrkwanController implements OrkwanTag
{
    /** Http request */
    public HttpServletRequest request;

    /** Http response */
    public HttpServletResponse response;

    /** Controller properties */
    public HashMap<String, String> properties = new HashMap<String, String>();

    /** Model data and objects */
    public HashMap<String, Object> model = new HashMap<String, Object>();

    /** Controller output type */
    public ControllerOutput output = ControllerOutput.UNDEFINED;

    /** Output xml document for output type DOCUMENT */
    public Document xmlDoc = DEFAULT_DOCUMENT;

    /** Output html source for output type SOURCE */
    public String htmlSource = DEFAULT_SOURCE;

    /** Invoked by URI */
    public boolean invokedByURI = false;

    /** Invoked by URL parameter */
    public boolean invokedByUrlParameter = false;

    /** Controller execution was successful */
    public boolean successful = false;

    /** Output encoding. If null, default encoding will be used */
    protected String encoding;

    /** content types */
    public static String CONTENTTYPE_HTML     = "text/html";
    public static String CONTENTTYPE_XML      = "text/xml";
    public static String DEFAULT_CHARSET_HTML = "utf-8";
    public static String DEFAULT_CHARSET_XML  = "utf-8";

    /** keys for intrinsic controller properties (mapped to OrkwanTags) */
    public static String PK_VIEW;
    public static String PK_VIEW_TYPE;
    public static String PK_URI;
    public static String PK_PARAMNAME;
    public static String PK_PARAMVALUE;

    /** request attribute key to controller */
    public static String RAK_CONTROLLER = "controller";

    /** default xml document */
    public static Document DEFAULT_DOCUMENT;

    /** default html source */
    public static String DEFAULT_SOURCE;

    static
    {
        initialize();
    }

    /**
     * Initialize class
     */
    private static void initialize()
    {
        // initialize property keys
        initializePropertyKeys();

        // create default xml document
        Element webapp = new Element("webapp");
        Element document = new Element("document");
        StringBuilder s = new StringBuilder("Default XML document in OrkwanController. ");
        s.append("This document was delivered since there is no document defined in your controller. ");
        s.append("To define an XML document call method setDocument(), or override method createDocument() in your controller.");
        document.setText(s.toString());
        webapp.addContent(document);
        DEFAULT_DOCUMENT = new Document(webapp);

        // create default html source
        s = new StringBuilder("<html><head>");
        s.append("<title>Default HTML source in OrkwanController</title>");
        s.append("</head><body>");
        s.append("Default HTML source in OrkwanController. ");
        s.append("This page was delivered since there is no html source defined in your controller. ");
        s.append("To define a html source call method setSource(), or override method createSource() in your controller.");
        s.append("</body></html>");
        DEFAULT_SOURCE = s.toString();
    }

    /**
     * initialize intrinsic property keys to same names as in OrkwanTags
     */
    private static void initializePropertyKeys()
    {
        PK_URI         = TAG_URI;        // uri pattern matching (may be null)
    	PK_VIEW        = TAG_VIEW;       // view output
        PK_VIEW_TYPE   = getPropertyAttributeKey(TAG_VIEW, ATTRIBUTE_TYPE); // view type (may be null)
        PK_PARAMNAME   = getPropertyAttributeKey(TAG_URLPARAMETER, ATTRIBUTE_NAME);  // url parameter name matching (may be null)
        PK_PARAMVALUE  = getPropertyAttributeKey(TAG_URLPARAMETER, ATTRIBUTE_VALUE); // url parameter value matching (may be null)
    }

    /**
     * Initialize controller
     */
    protected abstract void init() throws Exception;

    /**
     * Validate input parameters
     */
    protected abstract void validateInput() throws Exception;

    /**
     * Update model
     */
    protected abstract void updateModel() throws Exception;

    /**
     * Default constructor
     */
    public OrkwanController()
    {
    }

    /**
     * Get encoding.
     *
     * @return encoding The output encoding used by the controller
     */
    public String getEncoding()
    {
        return encoding;
    }

    /**
     * Set output encoding. If the output encoding is set, the content type is also set automatically.
     *
     * @see #setContentType()
     * @param encoding The new encoding
     */
    public void setEncoding(String encoding)
    {
        this.encoding = encoding;

        // http header content type must be set if encoding is changed
        setContentType();
    }

    /**
     * Get request.
     *
     * @return request The http request attached to the controller
     */
    public HttpServletRequest getRequest()
    {
        return request;
    }

    /**
     * Set http request
     *
     * @param request The http request used in the Controller
     */
    public void setRequest(HttpServletRequest request)
    {
        this.request = request;
    }

    /**
     * Get response.
     *
     * @return response The http response attached to the controller
     */
    public HttpServletResponse getResponse()
    {
        return response;
    }

    /**
     * Set http response
     *
     * @param response The http response used in the Controller
     */
    public void setResponse(HttpServletResponse response)
    {
        this.response = response;
    }

    /**
     * Get controller property.
     *
     * @param name Name of property
     * @return value Value of property
     */
    public String getControllerProperty(String name)
    {
        return properties.get(name);
    }

    /**
     * Set controller property.
     *
     * @param name Name of property
     * @param value Value of property
     */
    public void setControllerProperty(String name, String value)
    {
        properties.put(name, value);
    }

   /**
     * Get key to property attribute in properties map
    *
    * @param property The name of the property
    * @param attribute The name of the attribute
    */
    public static String getPropertyAttributeKey(String property, String attribute)
    {
        return property + ":" + attribute;
    }

    /**
     * Get controller property attribute.
     *
    * @param property The name of the property
    * @param attribute The name of the attribute
     * @return value Value of attribute
     */
    public String getControllerPropertyAttribute(String property, String attribute)
    {
        return properties.get(getPropertyAttributeKey(property, attribute));
    }

    /**
     * Get controller view
     *
     * @return view The controller view
     */
    public String getView()
    {
        return properties.get(PK_VIEW);
    }

    /**
     * Set controller view
     *
     * @param view The view of the controller
     */  
    public void setView(String view)
    {
        setControllerProperty(PK_VIEW, view);
    }

    /**
     * Get URI to which controller is matched (null if not matched)
     *
     * @return uri The controller uri
     */
    public String getUri()
    {
        return properties.get(PK_URI);
    }

    /**
     * Get Url parameter name to which controller is matched (null if not matched)
     *
     * @return url-parameter-name The controller url parameter name
     */
    public String getUrlParameterName()
    {
        return properties.get(PK_PARAMNAME);
    }

    /**
     * Get Url parameter value to which controller is matched (null if not matched)
     *
     * @return url-parameter-value The controller url parameter value
     */
    public String getUrlParameterValue()
    {
        return properties.get(PK_PARAMVALUE);
    }

    /**
     * Get value of view type property attribute
     *
     * @return viewtype The controller view type
     */
    public String getViewType()
    {
        return properties.get(PK_VIEW_TYPE);
    }

    /**
     * Get controller output type
     *
     * @return output The controller output type
     */
    public ControllerOutput getOutput()
    {
        return output;
    }

    /**
     * Set controller output type
     *
     * @param output The controller output type
     */
    public void setOutput(ControllerOutput output)
    {
        this.output = output;
    }

    /**
     * Get controller successful execution flag
     *
     * @return successful The controller successful flag
     */
    public boolean isSuccessful()
    {
        return successful;
    }

    /**
     * Set controller successful flag
     */
    public void setSuccessful()
    {
        this.successful = true;
    }

    /**
     * Get XML document
     *
     * @return xmlDoc The xml document
     */
    public Document getDocument()
    {
        return xmlDoc;
    }

    /**
     * Set XML document
     *
     * @param xmlDoc The xml document
     */
    public void setDocument(Document xmlDoc)
    {
        this.xmlDoc = xmlDoc;
    }

    /**
     * Get controller properties in map collection
     *
     * @return properties The controller properties
     */
    public HashMap<String, String> getProperties()
    {
        return properties;
    }

    /**
     * Get controller model map
     *
     * @return model The controller model map
     */
    public HashMap<String, Object> getModel()
    {
        return model;
    }

    /**
     * set controller model map
     *
     * @param model The model map
     */
    public void setModel(HashMap<String, Object> model)
    {
        this.model = model;
    }

    /**
     * Get controller invocation by URI flag
     *
     * @return invokedByURI The controller invocation by URI flag
     */
    public boolean isInvokedByURI()
    {
        return invokedByURI;
    }

    /**
     * Set controller invocation by URI flag
     *
     * @param invokedByURI The controller invocation by URI flag
     */
    public void setInvokedByURI(boolean invokedByURI)
    {
        this.invokedByURI = invokedByURI;
    }

    /**
     * Get controller invocation by URL parameter flag
     *
     * @return invokedByUrlParameter The controller invocation by URL parameter flag
     */
    public boolean isInvokedByUrlParameter()
    {
        return invokedByUrlParameter;
    }

    /**
     * Set controller invocation by URL parameter flag
     *
     * @param invokedByUrlParameter The controller invocation by URL parameter flag
     */
    public void setInvokedByUrlParameter(boolean invokedByUrlParameter)
    {
        this.invokedByUrlParameter = invokedByUrlParameter;
    }

    /**
     * Boot controller
     */
    protected void boot() throws Exception
    {
        // verify request
        if (request == null)
        {
            throw new Exception(getClass().getName() + " invoked without request defined (request is null)");
        }

        // verify if there already is a controller attached on the request (cascaded execution)
        OrkwanController prior = (OrkwanController) getAttribute(RAK_CONTROLLER);
        if (prior != null)
        {
            // copy model data from prior controller to this controller
            copyModel(prior);
        }

        // set this controller as attribute on request (used for creating view)
        setAttribute(RAK_CONTROLLER, this);

        // set response headers
        // Note: headers must be set early in execution flow in case the view is displayed early using e.g. createView()
        setResponseHeaders();
    }

    /**
     * Execute controller
     */
    public void execute() throws Exception
    {
        // boot controller
        boot();

        // init controller
        init();

        // validate input
        validateInput();

        // update model
        updateModel();

        // create output
        createOutput();

        // exit handling
        exit();

        // successful execution
        setSuccessful();
    }

    /**
     * Create output
     */
    protected void createOutput() throws Exception
    {
        // create document
        if (hasOutputDocument())
        {
            setDocument(createDocument());
        }

        // create source
        if (hasOutputSource())
        {
            setSource(createSource());
        }
    }

    /**
     * Set response headers. Override this method to set content types, cache directives etc.
     */
    protected void setResponseHeaders()
    {
        // by default only the content type is set
        setContentType();
    }

    /**
     * Set content type.
     *
     * Note: if the response writer has been created before this method is called, e.g. using response.getWriter(),
     * the character encoding for the response has already been set and this method will have no effect.
     */
    protected void setContentType()
    {
        if (response != null)
        {            
            switch (output)
            {
                case PAGE:
                case SOURCE:
                case CUSTOM:
                	
                    // set html content type
                    response.setContentType(getContentTypeDeclaration(CONTENTTYPE_HTML, encoding != null ? encoding : DEFAULT_CHARSET_HTML));
                    break;

                case DOCUMENT:

                    // set xml content type
                    response.setContentType(getContentTypeDeclaration(CONTENTTYPE_XML, encoding != null ? encoding : DEFAULT_CHARSET_XML));
                    break;

                default:
            }
        }
    }

    /**
     * Hook for exit handling. Override to customize exits. Default implementation is empty.
     */
    protected void exit() throws Exception
    {
        // defaults to no implementation
    }

    /**
     * Get content type declaration for http header
     * 
     * @return declaration The content type declaration
     */
    public static String getContentTypeDeclaration(String contentType, String charSet)
    {
        StringBuilder s = new StringBuilder(contentType);
        s.append(";charset=");
        s.append(charSet);
        return s.toString();
    }

    /**
     * Create view for this controller. Invoking this method will immediately abort the
     * execution flow in the controller and send the view to the client directly.
     */
    protected final void createView() throws Exception
    {
        throw new IncludeException(getView());
    }

    /**
     * Create specified view for this controller. Invoking this method will immediately abort the
     * execution flow in the controller and send the specified view to the client directly.
     * If this behavior is not desired, and the view shall be changed without
     * interupting the execution flow use the method setView()
     *
     * @see #setView(String)
     *
     * @param url Url to new view
     */
    protected final void createView(String url) throws Exception
    {
        throw new IncludeException(url);
    }
    
    /**
     * Forward specified view for this controller. Invoking this method will immediately abort the
     * execution flow in the controller and send the specified view to the client directly. It will
     * also change the URL of the request to that of the forward view.
     *
     * @see #setView(String)
     *
     * @param url Url to new view
     */
    protected final void forwardView(String url) throws Exception
    {
        throw new ForwardException(url);
    }

    /**
     * Redirect to url
     *
     * @param url The url to redirect to.
     */
    protected final void redirect(String url) throws Exception
    {
        throw new RedirectException(url);
    }

    /**
     * Create document. The default implementation does nothing and only returns the current xml document.
     * This method may be overriden in subclasses to provide a clear place holder for creation of
     * xml documents.
     *
     * @return doc The created xml document.
     */
    protected Document createDocument()
    {
        return getDocument();
    }

    /**
     * Get xml document as String
     * 
     * @return string the document as string
     */
    public String getXML()
    {
        if (encoding != null)
        {
            return XMLFactory.document2string(xmlDoc, true, encoding);
        }
        else
        {
            return XMLFactory.document2string(xmlDoc, true);
        }
    }

    /**
     * Default create html source
     * 
     * @return source The created html source
     */
    protected String createSource()
    {
        return getSource();
    }

    /**
     * Get html source
     * 
     * @return htmlSource The html source
     */
    public String getSource()
    {
        return htmlSource;
    }

    /**
     * Set html source
     * 
     * @param htmlSource The html source
     */
    public void setSource(String htmlSource)
    {
        this.htmlSource = htmlSource;
    }

    /**
     * Get request parameter
     * 
     * @param name The request parameter
     * @return value The value of the request parameter (may be null)
     */
    public String getParameter(String name)
    {
    	return name != null ? request.getParameter(name) : null;
    }

    /**
     * Get request attribute.
     *
     * @param name Attribute name
     * @return value Attribute value
     */
    public Object getAttribute(String name)
    {
        return request.getAttribute(name);
    }

    /**
     * Set request attribute
     *
     * @param name Attribute name
     * @param value Attribute value
     */
    public void setAttribute(String name, Object value)
    {
        request.setAttribute(name, value);
    }

    /**
     * Add cookie to response.
     *
     * @param cookieName name of cookie
     * @param cookieValue Value of cookie
     * @param maxAge Max age of cookie
     * @param domain Web domain
     * @param path Path
     */
    public void addCookie(String cookieName, String cookieValue, int maxAge, String domain, String path)
    {
        Cookie cookie = new Cookie(cookieName, cookieValue);
        cookie.setMaxAge(maxAge);
        cookie.setDomain(domain);
        cookie.setPath(path);
        response.addCookie(cookie);
    }

    /**
     * Get value of cookie in request (null if cookie not found)
     *
     * @param cookieName Name of cookie
     * @return cookieValue Value of cookie
     */
    public String getCookieValue(String cookieName)
    {
        if (cookieName != null && cookieName.length() > 0)
        {
            Cookie[] cookies = request.getCookies();
            if (cookies != null)
            {
                for (Cookie cookie : cookies)
                {
                    if (cookie.getName().equals(cookieName))
                    {
                        return cookie.getValue();
                    }
                }
            }
        }

        // cookie not found
        return null;
    }

    /**
     * Set object in model map. The object can later be retrieved in e.g. the controller or in a jsp.
     *
     * @param name Key to object in model map
     * @param value The object to set
     */
    public void setModelObject(String name, Object value)
    {
        model.put(name, value);
    }

    /**
     * Get model object from model map.
     *
     * @param name Key to object in model map
     * @return value The object in the map
     */
    public Object getModelObject(String name)
    {
        return name != null ? model.get(name) : null;
    }

    /**
     * Set String value in model map. The String can later be retrieved in e.g. the controller or in a jsp.
     * 
     * @param name Key to String value in model map
     * @param value The String to set
     */
    public void setModelData(String name, String value)
    {
        model.put(name, value);
    }

    /**
     * Get String from model map.
     *
     * @param name Key to String in model map
     * @return value The String in the map
     */
    public String getModelData(String name)
    {
        return (String) getModelObject(name);
    }

    /**
     * Set Boolean value in model map. The Boolean can later be retrieved in e.g. the controller or in a jsp.
     *
     * @param name Key to Boolean value in model map
     * @param value The Boolean to set
     */
    public void setModelBoolean(String name, boolean value)
    {
        model.put(name, value);
    }

    /**
     * Get Boolean from model map. If the Boolean is not in the map the value false is returned.
     *
     * @param name Key to Boolean in model map
     * @return value The Boolean in the map
     */
    public boolean getModelBoolean(String name)
    {
        Boolean b = (Boolean) getModelObject(name);
        return b != null && b;
    }

    /**
     * Set integer value in model map. The int can later be retrieved in e.g. the controller or in a jsp.
     *
     * @param name Key to int value in model map
     * @param value The int to set
     */
    public void setModelInt(String name, int value)
    {
        model.put(name, value);
    }

    /**
     * Get integer from model map. If the integer is not found 0 is returned.
     *
     * @param name Key to Integer in model map
     * @return value The Integer in the map
     */
    public int getModelInt(String name)
    {
        return (Integer) getModelObject(name);
    }

    /**
     * Get the url for the http request including the query string
     *
     * @return url The url to the request
     */
    public String getRequestURL() 
    {
        String query = request.getQueryString();
        String url = request.getRequestURL().toString();
        return (query != null) ? url + "?" + query : url;
    }

    /**
     * Get the uri for the request including the query string
     *
     * @return uri The uri to the request
     */
    public String getRequestURI()
    {
        String query = request.getQueryString();
        String uri = request.getRequestURI();
        return (query != null) ? uri + "?" + query : uri;
    }

    /**
     * Get the http session associated with the request
     *
     * @return session The http session associated with the request
     */
    public HttpSession getSession()
    {
        return request.getSession();
    }

    /**
     * Get attribute from http session
     *
     * @param name The name of the attribute
     * @return value The attribute stored in the http session
     */
    public Object getSessionAttribute(String name)
    {
        return getSession().getAttribute(name);
    }

    /**
     * Set attribute in http session.
     *
     * @param name The name of the attribute
     * @param value The value to store in the session
     */
    public void setSessionAttribute(String name, Object value)
    {
        getSession().setAttribute(name, value);
    }

   /**
     * Indicate if controller is configured to output a page (jsp or cascaded controller)
    *
    * @return page True if controller view is an url (jsp or cascaded servlet / controller)
     */
    public boolean hasOutputPage()
    {
        return output == ControllerOutput.PAGE;
    }

   /**
    * Indicate if controller is configured to output xml document.
    *
    * @return output True if controller has DOCUMENT (xml) output
    */
    public boolean hasOutputDocument()
    {
        return output == ControllerOutput.DOCUMENT;
    }

   /**
    * Indicate if controller is configured to output html source.
    *
    * @return output True if controller has SOURCE (html) output
    */
    public boolean hasOutputSource()
    {
        return output == ControllerOutput.SOURCE;
    }

   /**
    * Indicate if controller is configured to custom output.
    *
    * @return output True if controller has CUSTOM output
    */
    public boolean hasOutputCustom()
    {
        return output == ControllerOutput.CUSTOM;
    }

   /**
    * Copy configuration from another controller (template)
    *
    * @param template The controller to copy from
    */
    public void copy(OrkwanController template)
    {
        copyProperties(template);
        setOutput(template.getOutput());
        setInvokedByURI(template.isInvokedByURI());
        setInvokedByUrlParameter(template.isInvokedByUrlParameter());
    }

    /**
     * Copy properties from another controller
     *
     * @param controller The controller to copy from
     */
    public void copyProperties(OrkwanController controller)
    {
        for (String key : controller.getProperties().keySet())
        {
            setControllerProperty(key, controller.getProperties().get(key));
        }
    }

    /**
     * Copy model from another controller
     *
     * @param controller The controller to copy from
     */
    public void copyModel(OrkwanController controller)
    {
        for (String key : controller.getModel().keySet())
        {
            setModelObject(key, controller.getModel().get(key));
        }
    }

   /**
    * Configure controller.
    * The controller is configured after being loaded from xml file configuration.
    *
    * @param common Map of common properties
    */
    public void configure(HashMap<String, String> common) throws Exception
    {

       // set common properties
       setCommonProperties(common);

        // set output type
        setOutputType();

        // verify controller
        verify();
    }

   /**
    * configure controller based on common properties
    * 
    * @param common Map of common properties
    */
    private void setCommonProperties(HashMap<String, String> common)
    {
       // --- uri ---
       if (common.get(ATTRIBUTE_URI_ROOT) != null)
       {
           String uri = properties.get(PK_URI);

           // prepend uri-root only if uri is defined
           if (uri != null)
           {
               properties.put(PK_URI, common.get(ATTRIBUTE_URI_ROOT) + uri);
           }
       }

       // --- view ---
       if (common.get(ATTRIBUTE_VIEW_ROOT) != null)
       {
           String view = properties.get(PK_VIEW);

           // prepend view-root only if view is defined and does not start with a period
           // (a period indicates a relative adress)
           if (view != null && !view.startsWith("."))
           {
                properties.put(PK_VIEW, common.get(ATTRIBUTE_VIEW_ROOT) + view);
           }
       }
    }

   /**
    * Set controller output based on based on controller view property
    */
    private void setOutputType()
    {
       // default assignment is PAGE
       output = ControllerOutput.PAGE;

       String viewtype = getViewType();
        if (viewtype != null)
        {
            if (viewtype.equalsIgnoreCase(VALUE_PAGE))
            {
                output = ControllerOutput.PAGE;
            }
            else if (viewtype.equalsIgnoreCase(VALUE_XML))
            {
                output = ControllerOutput.DOCUMENT;
            }
            else if (viewtype.equalsIgnoreCase(VALUE_HTML))
            {
                output = ControllerOutput.SOURCE;
            }
            else if (viewtype.equalsIgnoreCase(VALUE_CUSTOM))
            {
                output = ControllerOutput.CUSTOM;
            }
            else if (viewtype.length() > 0)
            {
                // unrecognized view type
                output = ControllerOutput.UNDEFINED;
            }
        }

       if (output == ControllerOutput.PAGE)
       {
           String view = getView();
           if (view == null || view.length() == 0)
           {
               // page output is not defined
               output = ControllerOutput.UNDEFINED;
           }
       }
   }

    /**
     * Verify that controller is correctly configured
     */
    private void verify() throws Exception
    {
        if ((getUrlParameterName() == null || getUrlParameterValue() == null) && getUri() == null)
        {
            throw new Exception(getClass().getName() + " failed verification: could not be matched against a request URL");
        }

        if (output == ControllerOutput.UNDEFINED)
        {
            throw new Exception(getClass().getName() +  "failed verification: output is not defined or not recognized");
        }
    }

   /**
    * Get controller class (as defined in orkwan.xml file)
    *
    * @return class The class name
    */
    public String getControllerClass()
    {
        return getControllerProperty(TAG_CLASS);
    }

    /**
     * Query property of controller matching the specified url.
     * If a matching controller is not found null is returned.
     *
     * @param url String description of URL
     * @param property Name of property to query
     * @return value Value of property
     */
    public static String queryControllerProperty(String url, String property) throws Exception
    {
        if (url == null)
        {
            throw new Exception("The parameter url can not be null in controller query");
        }

        return Orkwan.getInstance().queryControllerProperty(url, property);
    }

    /**
     * Query view target of the controller matching the url.
     * If a matching controller is not found null is returned.
     *
     * @param url String description of URL
     * @return view The view of the controller
     */
    public static String queryControllerView(String url) throws Exception
    {
        if (url == null)
        {
            throw new Exception("The parameter url can not be null in controller query");
        }

        return Orkwan.getInstance().queryControllerView(url);
    }

    /**
     * Read request input stream and return result as an InputStreamReader. The stream will be read using the
     * encoding specified in the request (if specified).
     *
     * @return The input stream as an inputStreamReader for this request.
     */
    public InputStreamReader getInputStreamReader() throws IOException
    {        
        // get encoding from request (if specified)
        String encoding = request.getCharacterEncoding();
        boolean useEncoding = encoding != null;
        InputStreamReader inputStream = null;

        if (useEncoding)
        {
            try
            {
                inputStream = new InputStreamReader(request.getInputStream(), encoding);
            }
            catch (UnsupportedEncodingException use)
            {
                OrkwanLogger.error(this, "readRequest():  unsupported encoding = " + encoding + " , error = " + use);
                useEncoding = false;
            }
        }

        if (!useEncoding)
        {
            // use default reader
            inputStream = new InputStreamReader(request.getInputStream());
        }

        return inputStream;
    }

    /**
     * Read request input stream and return result as String. The stream will be read using the
     * encoding specified in the request (if specified).
     *
     * @param maxchar Maximum number of characters to read.
     * @return src The request as string
     */
    public String readRequest(int maxchar) throws IOException
    {
        InputStreamReader inputStreamReader = getInputStreamReader();

        // read the response
        BufferedReader reader = new BufferedReader(inputStreamReader);
        StringBuilder content = new StringBuilder();
        int c, index = 0;
        while ((c = reader.read()) != -1 && index++ <= maxchar)
        {
            content.append((char) c);
        }
        reader.close();

        return content.toString();
    }

    /**
     * Convert string to XML document. If conversion fails null is returned.
     *
     * @param xml String containing xml tags
     * @return document JDOM xml document
     */
    public Document string2document(String xml)
    {
        try
        {
            return XMLFactory.string2document(xml);
        }
        catch (Exception e)
        {
            // parse failed
            return null;
        }
    }

    /**
     * Convert xml document to string. If the controller has an output encoding specified
     * the xml document will include the encoding in the header.
     *
     * @param xml The JDOM xml document
     * @return string The output string
     */
    public String document2string(Document xml)
    {
        if (encoding != null)
        {
            return XMLFactory.document2string(xml, encoding);
        }
        else
        {
            return XMLFactory.document2string(xml);
        }
    }

    /**
     * Returns a string representation of the Controller.
     *
     * @return description The String desciption of the object
     */
    public String toString()
    {
        StringBuilder s = new StringBuilder();

        s.append("* Orkwan: ").append(getClass().getName());
        if (getUri() != null)
        {
            s.append(", uri =").append(getUri());
        }
        if (getUrlParameterName() != null && getUrlParameterValue() != null)
        {
            s.append(", url-param: ").append(getUrlParameterName()).append("=").append(getUrlParameterValue());
        }
        s.append(", output=").append(getOutput());
        if (hasOutputPage())
        {
             s.append(", view=").append(getView());
        }

        return s.toString();
    }

    /**
     * Print a oneliner about this controller.
     */
    public void print()
    {
        System.out.println(toString());
    }

    /**
     * Print controller properties.
     */
    public void printProperties()
    {
        StringBuilder sb = new StringBuilder("Controller properties:\n");
        for (String name : properties.keySet()) {
            sb.append("property = ").append(name).append(", value = ").append(getControllerProperty(name)).append("\n");
        }

        System.out.println(sb.toString());
    }
}
