/**
 This file is software from the ORKWAN project.
 http://www.orkwan.com
 
 Copyright (C) 2000-2008 Henrik Nyberg.
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions, and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions, and the disclaimer that follows
    these conditions in the documentation and/or other materials
    provided with the distribution.

 3. The name "ORKWAN" must not be used to endorse or promote products
    derived from this software without prior written permission.  For
    written permission, please contact <info_AT_orkwan_DOT_com>.

 4. Products derived from this software may not be called "ORKWAN", nor
    may "ORKWAN" appear in their name, without prior written permission
    from the ORKWAN Project <info_AT_orkwan_DOT_com>.

 In addition, we request (but do not require) that you include in the
 end-user documentation provided with the redistribution and/or in the
 software itself an acknowledgement equivalent to the following:
     "This product includes software developed by the
      ORKWAN Project (http://www.orkwan.com/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR THE PROJECT
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
 */

package com.orkwan;

import javax.servlet.RequestDispatcher;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.PrintWriter;

/**
 * OrkwanServlet is the abstract base class servlet for a control servlet. OrkwanServlet manages the entire
 * web control flow by routing the http request to a matching controller, executing the controller and
 * creating the http response to the remote client.
 *
 * Subclass this servlet to create a control servlet in a web application. Mount the control servlet
 * to the web context for the application in the web.xml file.
 *
 * @author Henrik Nyberg
 */
public abstract class OrkwanServlet extends HttpServlet
{
    // counters
    public int countProcess = 0;
    public int countExecute = 0;
    public int countRedirect = 0;
    public int countForward = 0;
    public int countInclude = 0;
    public int countWrite = 0;
    public int countErrorIO = 0;
    public int countErrorServlet = 0;
    public int countError = 0;
    
    /**
     * constructor
     */
    public OrkwanServlet()
    {
        super();
    }

    /**
     * Initializes the servlet and logs the initialization by calling super.init(config).
     *
     * @param config Servlet configuration information.
     */
    public void init(ServletConfig config) throws ServletException
    {
        super.init(config);
        
        Orkwan.getInstance().registerServlet(this);
    }

    /**
     * Performs the HTTP GET operation.
     *
     * @param request  an object that encapsulates the request to the servlet.
     * @param response an object that encapsulates the response from the servlet.
     * @throws ServletException if the request could not be handled
     * @throws java.io.IOException if the request resulted in I/O problems
     */
    public void doGet(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException
    {
        process(request, response);
    }

    /**
     * Performs the HTTP POST operation.
     *
     * @param request  an object that encapsulates the request to the servlet.
     * @param response an object that encapsulates the response from the servlet.
     * @throws ServletException if the request could not be handled.
     * @throws IOException if the request resulted in I/O problems.
     */
    public void doPost(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException
    {
        process(request, response);
    }

    /**
     * Perform the standard request processing for this request, and create the corresponding response.
     *
     * @param request  The servlet request we are processing
     * @param response The servlet response we are creating
     * @throws IOException if an input/output error occurs
     * @throws ServletException if a servlet exception is thrown
     */
    protected void process(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException
    {
        countProcess++;
        OrkwanController controller = null;

        try
        {
            // init processing
            init(request, response);

            // perform validation
            validate(request, response);
            
            // load controller
            controller = load(request, response);
            
            if (controller != null)
            {
            	// execute controller
            	execute(request, response, controller);
            }
            else
            {
            	// handle not found
            	handleNotFound(request, response);
            }
        }
        catch (RedirectException eRedirect)
        {
            redirect(request, response, eRedirect.getUrl());
        }
        catch (ForwardException eForward)
        {
            forward(request, response, eForward.getUrl());
        }
        catch (IncludeException eInclude)
        {
            include(request, response, eInclude.getUrl());
        }
        catch (IOException eIO)
        {
            countErrorIO++;
            OrkwanLogger.error(this, "IO Exception: " + eIO, eIO);
            throw new IOException(eIO.getMessage());
        }
        catch (ServletException eSvt)
        {
            countErrorServlet++;
            OrkwanLogger.error(this, "Servlet Exception: " + eSvt, eSvt);
            throw new ServletException(eSvt.getMessage());
        }
        catch (Throwable t)
        {
            countError++;
            t.fillInStackTrace();
            
            if (controller != null)
            {
                OrkwanLogger.error(this, "Runtime Error in " + controller.getClass(), t);
                Orkwan.getInstance().getControllerMetric(controller).error(t);           	
            }
            else
            {
            	OrkwanLogger.error(this, "Runtime Error", t);
            }
            handleError(request, response, t, controller);
        }
    }

    /**
     * Initialize processing
     *
     * @param request  The servlet request we are processing
     * @param response The servlet response we are creating
     */
    protected abstract void init(HttpServletRequest request, HttpServletResponse response) throws Exception;

    /**
     * Perform validation. Use this method to verify cookies, session etc.
     *
     * @param request  The servlet request we are processing
     * @param response The servlet response we are creating
     */
    protected abstract void validate(HttpServletRequest request, HttpServletResponse response) throws Exception;

    /**
     * Handle not found. This method is called when there is no controller matching the http request
     *
     * @param request  The servlet request we are processing
     * @param response The servlet response we are creating
     */
    protected abstract void handleNotFound(HttpServletRequest request, HttpServletResponse response) throws Exception;
  
    /**
     * Handle error. This method is called when an exception is thrown during execution. The controller may be null.
     *
     * @param request  The servlet request we are processing
     * @param response The servlet response we are creating
     * @param t  The error throwable
     */
    protected abstract void handleError(HttpServletRequest request, HttpServletResponse response, Throwable t, OrkwanController controller) throws IOException, ServletException;

    /**
     * Load Controller
     *
     * @param request  The servlet request we are processing
     * @param response The servlet response we are creating
     */
    protected OrkwanController load(HttpServletRequest request, HttpServletResponse response) throws Exception
    {
    	OrkwanController controller = Orkwan.getInstance().loadController(request, response);
    	
        if (controller != null)
        {
        	Orkwan.getInstance().getControllerMetric(controller).load(controller.isInvokedByURI());
        }
        
        return controller;
    }
    
    /**
     * Execute Controller
     *
     * @param request  The servlet request we are processing
     * @param response The servlet response we are creating
     */
    protected void execute(HttpServletRequest request, HttpServletResponse response, OrkwanController controller) throws Exception
    {
        countExecute++;  
        long start = System.currentTimeMillis();
        
        // --- start of execution ---
        
    	controller.execute();
        output(request, response, controller);
        
        // --- end of execution ---
        
    	long stop = System.currentTimeMillis();
    	Orkwan.getInstance().getControllerMetric(controller).execute(stop - start);
    }

    /**
     * Create output for Controller
     *
     * @param request  The servlet request we are processing
     * @param response The servlet response we are creating
     * @param controller The Orkwan controller
     */
    protected void output(HttpServletRequest request, HttpServletResponse response, OrkwanController controller) throws Exception
    {
        switch (controller.output)
        {
            case PAGE:

                if (controller.isInvokedByURI())
                {
                    // forward to view when invoked by URI
                    forward(request, response, controller.getView());
                }
                else
                {
                    // include view when invoked by url parameter
                    include(request, response, controller.getView());
                }

                break;

            case DOCUMENT:

                // write xml document to response
                write(response, controller.getXML());

                break;

            case SOURCE:

                // write html source to response
                write(response, controller.getSource());

                break;

            case CUSTOM:

                // Do nothing the output should already have been written to the response.
                break;

            default:

                 // controller has no configured output
                write(response, "Orkwan: No output configured");
        }
    }

    /**
     * Forward request.
     *
     * @param request  The servlet request we are processing
     * @param response The servlet response we are creating
     * @param url The url to forward to
     */
    protected void forward(HttpServletRequest request, HttpServletResponse response, String url) throws IOException, ServletException
    {
        countForward++;

        RequestDispatcher rd = request.getRequestDispatcher(url);

        if (rd == null)
        {
            response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
        }
        else
        {
            rd.forward(request, response);
        }
    }

    /**
     * Include request.
     *
     * @param request  The servlet request we are processing
     * @param response The servlet response we are creating
     * @param url      The url to include
     */
    protected void include(HttpServletRequest request, HttpServletResponse response, String url) throws IOException, ServletException
    {
        countInclude++;

        RequestDispatcher rd = request.getRequestDispatcher(url);

        if (rd == null)
        {
            response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
        }
        else
        {
            rd.include(request, response);
        }
    }

    /**
     * Write directly to response.
     *
     * @param response The servlet response we are creating
     * @param content  The content we are writing
     */
    protected void write(HttpServletResponse response, String content) throws IOException
    {
        countWrite++;

        /*
            By default, the PrintWriter uses the ISO-8859-1 (Latin-1) charset, appropriate for most Western European languages.
            To specify an alternate charset, the charset value must be passed to the setContentType() method before
            the servlet retrieves its PrintWriter. Override the method OrkwanController.setResponseHeaders(HttpServletResponse response)
            to set a custom content type for the PrintWriter.
        */

        PrintWriter writer = response.getWriter();
        writer.write(content);
        writer.flush();
    }

    /**
     * Redirect request.
     *
     * @param request  The servlet request we are processing
     * @param response The servlet response we are creating
     * @param url      The url we are redirecting to
     */
    protected void redirect(HttpServletRequest request, HttpServletResponse response, String url)
    {
        countRedirect++;

        try
        {
            // redirect current browser window to new url
            response.sendRedirect(url);
        }
        catch (IllegalStateException e)
        {
            OrkwanLogger.error(this, "Redirect on committed response: url =" + url);
        }
        catch (Exception e)
        {
        	OrkwanLogger.error(this, "Redirect failed: url =" + url);
        }

        // note: if redirect failed do nothing, since response is already committed or there is an I/O error
    }

	/**
	 * @return the countProcess
	 */
	public int getCountProcess()
	{
		return countProcess;
	}

	/**
	 * @return the countError
	 */
	public int getCountError()
	{
		return countError;
	}

	/**
	 * @return the countErrorIO
	 */
	public int getCountErrorIO()
	{
		return countErrorIO;
	}

	/**
	 * @return the countErrorServlet
	 */
	public int getCountErrorServlet()
	{
		return countErrorServlet;
	}

	/**
	 * @return the countExecute
	 */
	public int getCountExecute()
	{
		return countExecute;
	}

	/**
	 * @return the countForward
	 */
	public int getCountForward()
	{
		return countForward;
	}

	/**
	 * @return the countInclude
	 */
	public int getCountInclude()
	{
		return countInclude;
	}

	/**
	 * @return the countRedirect
	 */
	public int getCountRedirect()
	{
		return countRedirect;
	}

	/**
	 * @return the countWrite
	 */
	public int getCountWrite()
	{
		return countWrite;
	}

}
